/*
    FUNIONFS: UNIONFS over FUSE Usermode filesystem
    Copyright (C) 2005-2006  Stephane APIOU <stephane.apiou@free.fr>

    This program can be distributed under the terms of the GNU GPL.
    See the file COPYING.
*/

/* A FAIRE
 * 
 * Faire le TODO
 *
 * A faire: pour les liens symboliques, la creation de lien doit tenir compte des chemins relatifs
 * de la version en lecture seule /  l'arborescence monte pour des liens pointant en dehors du
 * systme unionfs et elle doit tenir compte des pointages des liens des arborescences en lecture seule
 * et en lecture/ecriture. 
 * Cette fonctionnalit doit tre optionnelle
 */

// use the new API

#define VERSION "0.4.1"

#include <fuse.h>
#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <dirent.h>
#include <sys/mman.h>
#include <utime.h>

#if FUSE_VERSION < 25
#include "compat/fuse_opt.h"
#endif

#ifdef HAVE_SETXATTR
	#include <sys/xattr.h>
#endif

#define MAX_LINKPATH	4096
#define MAX_READLINKS	50

#define DELETED_STRING "_DELETED~"

#define FILE_EXIST	0X01
#define FILE_WRITEABLE	0X02
#define FILE_DELETED	0X04
#define FILE_CREATABLE	0X08
#define FILE_TODELETE	0X10

#define IDX_GET	0
#define IDX_PUT	1
#define IDX_NEW	2

#define _DEBUG	1

#if _DEBUG
	#define DEBUG_PRINTF(args...) \
		do { if (funionfs.debug) fprintf(stderr, args); } while(0)
#else
	#define DEBUG_PRINTF(args...)
#endif


struct unionfs_desc {
	char *path;
	char	ro;
	struct unionfs_desc *pnext;
};

struct unionfs_walk {
	char *path;
	char *str;
	struct unionfs_desc *curdir;
};

unsigned long maxpath;
struct unionfs_desc *unionfs_list;

struct funionfs {
	char 	*uid;
	char 	*gid;
	char 	*mode;
	int	delete;
	int	copyup;
	int	debug;
	char 	*dirlist;
	char	*firstdir;
	char	*mountpoint;
	char	*del_string;
	int	del_len;
};

struct funionfs funionfs = { "","","", 0, 0, 0, "", "", "", DELETED_STRING, 0};

int getrightstr(char *str, char *substr)
{
	char *pend, *psend;

	pend = str + strlen(str)-1;
	psend = substr + strlen(substr)-1;

	if (strlen (substr) > strlen(str))
		return 0;

	while (psend >= substr) {
		if (*pend != *psend)
			break;
		pend--;
		psend--;
	}
	if (psend < substr)
		return strlen(substr);
	return 0;
}
	
int issubpath(char *abspath, char *basepath)
{
	if (*basepath != '/')
		return 0;

	if (*abspath != '/')
		return 0;

	while (*++abspath == *++basepath) {
		/* abspath and base path are equal */
		if (*abspath == '\0')
			return 1;
	}
	if (*abspath == '\0')
		return 1;
	return 0;
}

char *abs2relpath(char *abspath, char *basepath)
{
	char *result;

	if (*basepath != '/')
		return NULL;

	if (*abspath != '/')
		return NULL;

	while (*++abspath == *++basepath) {
		/* abspath and base path are equal */
		if (*abspath == '\0')
			return NULL;
	}
	/* return just after the last / */
	while (abspath[-1] != '/')
		abspath--;
	while (basepath[-1] != '/')
		basepath--;

	result = (char *)malloc(strlen(basepath)*2 + strlen(abspath) + 1);
	if (result == NULL)
		return NULL;

	while(*basepath !='\0') {
		if (*basepath == '/')
			strcat(result,"../");
	}
	if (basepath[-1] != '/')
		strcat(result,"../");
	strcat(result, abspath);

	return result;
}

/* warning basepath must start with a slash */

char *rel2abspath(char *relpath, char *basepath)
{
	char *copy_path;
	char *copy_bpath;
	char *orig_cpypath;
	char *temp_path;
	char link_path[MAX_LINKPATH];
	char *new_path;
	char *max_path;
	int readlinks = 0;
	int n;

	/* test if basepath is an absolute path */

	if (*basepath != '/')
		return NULL;

	/* Make a copy of the source path since we may need to modify it. */
	copy_path = strdup(relpath);
	if (copy_path == NULL)
		return NULL;
	orig_cpypath = copy_path;

	/*
	 * Make a copy of basepath and allocate a buffer of sizeof(relpath)+sizeof(basepath) since
	 * the resulting path can't be larger
	 */
	copy_bpath = (char *)malloc(strlen(basepath)+strlen(relpath)+3);
	if (copy_bpath == NULL) {
		free(orig_cpypath);
		return NULL;
	}
	strcpy(copy_bpath, basepath);
	new_path = copy_bpath;
	max_path = copy_bpath +strlen(basepath)+strlen(relpath);

	/* If it's a relative pathname use getwd for starters. */
	if (*copy_path != '/') {
		new_path += strlen(new_path);
		if (new_path[-1] != '/')
			*new_path++ = '/';
	}
	else {
		*new_path++ = '/';
		copy_path++;
	}
	/* Expand each slash-separated pathname component. */
	while (*copy_path != '\0') {
		/* Ignore stray "/". */
		if (*copy_path == '/') {
			copy_path++;
			continue;
		}
		if (*copy_path == '.') {
			/* Ignore ".". */
			if (copy_path[1] == '\0' || copy_path[1] == '/') {
				copy_path++;
				continue;
			}
			if (copy_path[1] == '.') {
				if (copy_path[2] == '\0' || copy_path[2] == '/') {
					copy_path += 2;
					/* Ignore ".." at root. */
					if (new_path == copy_bpath + 1)
						continue;
					/* Handle ".." by backing up. */
					while ((--new_path)[-1] != '/')
						;
					continue;
				}
			}
		}
		/* Safely copy the next pathname component. */
		while (*copy_path != '\0' && *copy_path != '/') {
			if (new_path > max_path) {
				free(copy_bpath);
				free(orig_cpypath);
				return NULL;
			}
			*new_path++ = *copy_path++;
		}

		/* Protect against infinite loops. */
		if (readlinks++ > MAX_READLINKS) {
			free(copy_bpath);
			free(orig_cpypath);
			return NULL;
		}
		/* See if latest pathname component is a symlink. */
		*new_path = '\0';
		n = readlink(copy_bpath, link_path, MAX_LINKPATH - 1);
		if (n < 0) {
			/* EINVAL means the file exists but isn't a symlink. */
			if (errno != EINVAL) {
				free(copy_bpath);
				free(orig_cpypath);
				return NULL;
			}
		} else {
			/* Note: readlink doesn't add the null byte. */
			link_path[n] = '\0';
			if (*link_path == '/') {
				/* Start over for an absolute symlink. */
				*copy_bpath='\0';
			} else {
				/* Otherwise back up over this component. */
				while (*(--new_path) != '/')
					;
				new_path[1]='\0';
			}
			temp_path = (char *)malloc(strlen(copy_bpath)+strlen(copy_path)+strlen(link_path)+4);
			if (temp_path == NULL) {
				free(copy_bpath);
				free(orig_cpypath);
				return NULL;
			}
			strcpy(temp_path, copy_bpath);
			free(copy_bpath);
			copy_bpath = temp_path;
			new_path = copy_bpath;
			max_path = copy_bpath +strlen(copy_bpath)+strlen(copy_path)+strlen(link_path)+3;

			/* Insert symlink contents into copy_path. */
			temp_path = (char *)malloc(strlen(copy_path)+strlen(link_path)+1);
			if (temp_path == NULL) {
				free(copy_bpath);
				free(orig_cpypath);
				return NULL;
			}
			strcpy(temp_path, link_path);
			strcat(temp_path, copy_path);
			free(orig_cpypath);
			orig_cpypath = temp_path;
			copy_path = temp_path;
		}
		*new_path++ = '/';
	}
	/* Delete trailing slash but don't whomp a lone slash. */
	if (new_path != copy_bpath + 1 && new_path[-1] == '/')
		new_path--;
	/* Make sure it's null terminated. */
	*new_path = '\0';

	temp_path = strdup(copy_bpath);
	if (temp_path == NULL) {
		free(copy_bpath);
		free(orig_cpypath);
		return NULL;
	}
	free(copy_bpath);
	free(orig_cpypath);

	return temp_path;
}

int funionfs_addpath(char *path, int ro)
{
	struct unionfs_desc *pdesc, *plast;
	char *copy_path;
	int res;
	struct stat statfile;

	res = lstat(path, &statfile);
	if (res < 0)
		return -errno;
	if (strlen(path) > maxpath)
		maxpath = strlen(path);

	copy_path = strdup(path);
	if (copy_path == NULL)
		return (-ENOMEM);

	pdesc = (struct unionfs_desc *)malloc(sizeof(struct unionfs_desc));
	if (pdesc == NULL) {
		free(copy_path);
		return (-ENOMEM);
	}
	if (unionfs_list != NULL) {
		plast = unionfs_list;
		while(plast->pnext != NULL)
			plast = plast->pnext;
	}

	pdesc->path = copy_path;
	pdesc->ro = ro;
	pdesc->pnext = NULL;
	if (unionfs_list != NULL)
		plast->pnext = pdesc;
	else
		unionfs_list = pdesc;

	return 1;

}

/* get path for read and write 
 * return is:
 *  negative: error (-errno)
 *  bits are as follow up
 *  0123
 *  ^^^^
 *  ||||
 *  |||\_file can be created in this path
 *  ||\_ file is delete in write branch
 *  |\__ file exist in write branch
 *  \___ file exist in read branch
 */

int funionfs_realpath(const char *path, char **get_path, char **put_path, char **new_path, struct unionfs_desc *tab[3])
{
	int len;
	struct unionfs_desc *pdesc, *pnew;
	int res;
	char *cur_path, *pend;
	struct stat statfile;
	int deleted, to_delete;

	len = strlen(path) + maxpath + funionfs.del_len + 1;
	cur_path = (char *)malloc(len);
	if(cur_path == NULL)
		return -ENOMEM;

	*get_path = NULL;
	*put_path = NULL;
	*new_path = NULL;
	deleted = 0;
	to_delete = 0;
	pnew=NULL;

	for(pdesc = unionfs_list; pdesc != NULL;pdesc=pdesc->pnext)
	{
		strncpy(cur_path,pdesc->path, len);
		pend = cur_path + strlen(cur_path);
		while (*--pend == '/') ;
		pend[1]='\0';
		strncat(cur_path,path, len);
		/* delete trailing slashs */
		pend = cur_path + strlen(cur_path);
		while (*--pend == '/') ;
		pend[1]='\0';

		res = 0;
		DEBUG_PRINTF("%s\n", cur_path);
		if(lstat(cur_path, &statfile) >= 0)
			res = 1;

		strcat(cur_path, funionfs.del_string);
		DEBUG_PRINTF("%s\n", cur_path);
		if(lstat(cur_path, &statfile) >= 0)
			res = 2;
		pend[1]='\0';

		switch(res)
		{
		case 2:
			deleted=1;
			DEBUG_PRINTF("DELETED\n");
			if (*get_path != NULL)
				free(*get_path);
			*get_path=NULL;
			if (*put_path != NULL)
				free(*put_path);
			*put_path=NULL;
			break;
		case 1:
			if (*get_path != NULL) {
				DEBUG_PRINTF("TO DELETE\n");
				to_delete=1;
			}
			if (pdesc->ro) {
				if (*get_path != NULL)
					free(*get_path);
				*get_path = cur_path;
				tab[IDX_GET] = pdesc;
			} else {
				if (*get_path != NULL)
					free(*get_path);
				*get_path = cur_path;
				tab[IDX_GET] = pdesc;
				cur_path = (char *)malloc(len);
				if(cur_path == NULL) {
					free(*get_path);
					return -ENOMEM;
				}
				strcpy(cur_path, *get_path);
				*put_path = cur_path;
				tab[IDX_PUT] = pdesc;
			}
			cur_path = (char *)malloc(len);
			if(cur_path == NULL) {
				if (*get_path != NULL)
					free(*get_path);
				if (*put_path != NULL)
					free(*put_path);
				return -ENOMEM;
			}
			break;
		}
		if (! pdesc->ro)
			pnew=pdesc;
	}

	if (pnew !=  NULL) {
		*new_path = cur_path;
		strncpy(cur_path,pnew->path, len);
		pend = cur_path + strlen(cur_path);
		while (*--pend == '/') ;
		pend[1]='\0';
		strncat(cur_path,path, len);
		/* delete trailing slashs */
		pend = cur_path + strlen(cur_path);
		while (*--pend == '/') ;
		pend[1]='\0';
		tab[IDX_NEW] = pnew;
	} else
		free(cur_path);
	
	res =(*put_path != NULL) << 1 | (*get_path != NULL) | deleted << 2 | (*new_path != NULL) << 3 | to_delete << 4;

	if (*put_path == NULL) {
		*put_path=(char *)malloc(1);
		**put_path = '\0';
	}
	if (*get_path == NULL) {
		*get_path=(char *)malloc(1);
		**get_path = '\0';
	}
	if (*new_path == NULL) {
		*new_path=(char *)malloc(1);
		**new_path = '\0';
	}
	return res;
}

struct unionfs_walk *funionfs_openpath(const char *path)
{
	struct unionfs_walk *pwalk;

	pwalk = (struct unionfs_walk*)malloc( sizeof(struct unionfs_walk));
	if (pwalk == NULL) {
		errno = ENOMEM;
		return NULL;
	}
	pwalk->path = strdup(path);
	if (pwalk == NULL) {
		free(pwalk);
		errno = ENOMEM;
		return NULL;
	}
	pwalk->curdir = unionfs_list;
	pwalk->str = NULL;

	return pwalk;
}

char *funionfs_readpath(struct unionfs_walk *pwalk, long *type)
{
	struct unionfs_desc *pdesc;
	char *str;
	char *pend;
	int res;
	struct stat statfile;

	while (1) {
		if (pwalk->str != NULL) {
			free(pwalk->str);
			pwalk->str = NULL;
		}

		pdesc = pwalk->curdir;

		if (pdesc == NULL) {
			*type = 0;
			return NULL;
		}
		pwalk->curdir=pdesc->pnext;

		str = (char *)malloc(strlen(pdesc->path) + strlen(pwalk->path) + 1);
		if (str == NULL) {
			*type = 0;
			errno = ENOMEM;
			return NULL;
		}
		pwalk->str = str;

		strcpy(str,pdesc->path);

		/* delete trailing slashs */
		pend = str + strlen(str);
		while (*--pend == '/') ;
		pend[1]='\0';

		strcat(str,pwalk->path);

		/* delete trailing slashs */
		pend = str + strlen(str);
		while (*--pend == '/') ;
		pend[1]='\0';

		res = 0;
		DEBUG_PRINTF("%s\n", str);
		if(lstat(str, &statfile) >= 0)
			res = 1;
		if (res)
		{
			if (pdesc->ro) {
				*type = 1;
			} else {
				*type = 3;
			}
			return str;
		}
	}

	return NULL;
}

int funionfs_closepath(struct unionfs_walk *pwalk)
{
	if (pwalk != NULL) {
		if (pwalk->str != NULL)
			free(pwalk->str);
		if (pwalk->path != NULL)
			free(pwalk->path);
		free(pwalk);
	}
	return 1;
}

int funionfs_copytowrite(char *readpath, char *writepath)
{
	char *wstr, *rstr;
	char *readfile, *writefile;
	char *wpend, *rpend;
	struct stat sb, rsb;
	int readfd, writefd;
	char link_path[MAX_LINKPATH];
	struct utimbuf buf;
	int endflag;

	/* create the path to file */
	wstr = strdup(writepath);
	if(wstr == NULL)
	{
		errno=ENOMEM;
		return -1;
	}

	rstr = strdup(readpath);
	if(rstr == NULL)
	{
		free(wstr);
		errno=ENOMEM;
		return -1;
	}
#if 0
	/* detect last element of pathname */
	while ((*--wpend != '/') && (wpend >= wstr)) ;
	wpend[0]='\0';

	/* detect last element of pathname */
	rpend = rstr + strlen(rstr);
	while ((*--rpend != '/') && (rpend >= rstr)) ;
	rpend[0]='\0';
#endif
	wpend = wstr + strlen(wstr);
	rpend = rstr + strlen(rstr);

	/* find start of variable element of both pathnames */
	while ((*wpend == *rpend) && (wpend >= wstr) && (rpend >= rstr)) {
		wpend --;
		rpend --;
	}
	while((*wpend != '/') && (*wpend != '\0'))
		wpend++;
	while((*rpend != '/') && (*rpend != '\0'))
		rpend++;
	endflag=-1;

	/* walk thru the path and create each element to the last element*/
	/* if the element does'nt exist, clone it from the readonly path */

	while(endflag < 0) {
		/* suppress leading slash */
		while(*wpend == '/')
			wpend++;
		while(*rpend == '/')
			rpend++;
		/* detect pathname element */
		while ((*wpend != '/') && (*wpend != '\0'))
			wpend++;
		if (*wpend == '\0')
			endflag=1;
		*wpend='\0';
		while ((*rpend != '/') && (*rpend != '\0'))
			rpend++;
		if (*rpend == '\0')
			endflag=1;
		*rpend='\0';

		/* get element attributes from the read path */
		if(lstat(rstr, &rsb) < 0) {
			break;
		}
		/* does the file exist in the writing path ? */
		if(lstat(wstr, &sb) < 0)
		{
			if (S_ISDIR(rsb.st_mode)) { /* a directory ? : create it */
				if (mkdir(wstr, rsb.st_mode & 07777) < 0) 
					break;
			}
			if (S_ISSOCK(rsb.st_mode)||S_ISCHR(rsb.st_mode)||S_ISBLK(rsb.st_mode)||S_ISFIFO(rsb.st_mode)) {
				/* a device or a pipe ? : create it */
				if (mknod(wstr, rsb.st_mode, rsb.st_rdev) < 0)
					break;
			}
			if (S_ISLNK(rsb.st_mode)) { /* a link ?:create it */
				if (readlink(readpath, link_path, MAX_LINKPATH - 1) < 0)
					break;
				if (symlink(wstr, link_path) < 0)
					break;
			}
			if (S_ISREG(rsb.st_mode)) { /* a file ?: copy it */
				if ((readfd = open (readpath, O_RDONLY)) < 0)
					break;

				if ((writefd = open (wstr, O_CREAT | O_RDWR, S_IRWXU)) < 0)
				{
					close(readfd);
					break;
				}

				if(ftruncate(writefd, rsb.st_size) < 0)
				{
					close(readfd);
					close(writefd);
					break;
				}

				if ((readfile = (char *) mmap(NULL, rsb.st_size, PROT_READ, MAP_SHARED, readfd, 0)) == (char *)-1)
				{
					close(readfd);
					close(writefd);
					break;
				}

				if ((writefile = (char *) mmap(NULL, rsb.st_size, PROT_WRITE, MAP_SHARED, writefd, 0)) == (char *)-1)
				{
					close(readfd);
					close(writefd);
					break;
				}

				memcpy(writefile, readfile, rsb.st_size);

				if(munmap(readfile, rsb.st_size) < 0)
				{
					close(readfd);
					close(writefd);
					break;
				}

				if(munmap(writefile, rsb.st_size) < 0)
				{
					close(readfd);
					close(writefd);
					break;
				}
				close(readfd);
				close(writefd);
			}
			/* set permissions, group, user and time of the readonly file */
			if (chmod(wstr, rsb.st_mode & 07777) < 0)
				break;
			if (chown(wstr, rsb.st_uid, rsb.st_gid) < 0)
				break;
			buf.actime = rsb.st_atime;
			buf.modtime = rsb.st_mtime;
			if (utime(wstr, &buf) < 0)
				break;
		}
		if (endflag > 0)
			break;
		*wpend='/';
		*rpend='/';
	}
	free(wstr);
	free(rstr);

	return endflag;
}


int funionfs_pathtowrite(const char *path)
{
	struct unionfs_desc *pdesc_tab[3];

	char *getpath, *setpath, *newpath;
	char *str, *pend;
	int res;

	/* create the path to file */
	str = strdup(path);
	if(str == NULL)
	{
		errno=ENOMEM;
		return -1;
	}

	/* detect last element of pathname */
	pend = str + strlen(str);
	while ((*--pend != '/') && (pend >= str)) ;
	pend[0]='\0';

	res =funionfs_realpath(str, &getpath, &setpath, &newpath, pdesc_tab);
	if (res < 0) {
		free(str);
		return res;
	}


	if (*setpath == '\0') {
		res = funionfs_copytowrite(getpath, newpath);
	} else
		res=0;

	free(str);
	free(getpath);
	free(setpath);
	free(newpath);

	if (res < 0) 
		return -errno;

	return 0;
}

static int funionfs_getattr(const char *path, struct stat *stbuf)
{
	int res = 0;
	struct unionfs_desc *pdesc_tab[3];

	char *getpath, *setpath, *newpath;
	struct stat statfile;

	res =funionfs_realpath(path, &getpath, &setpath, &newpath, pdesc_tab);
	if (res < 0)
		return res;

	DEBUG_PRINTF("GETATTR(%s,%s,%s)\n", getpath, setpath, newpath);

	if (res & FILE_EXIST) {
		res = lstat(getpath, &statfile);
	} else {
		res = -1;
		errno=ENOENT;
	}

	free(getpath);
	free(setpath);
	free(newpath);

	if (res < 0) 
		return -errno;

	memcpy(stbuf, &statfile, sizeof(struct stat));

	return 0;
}

static int funionfs_readlink(const char *path, char *buf, size_t size)
{
	char *getpath, *setpath, *newpath;
	int res;
	struct unionfs_desc *pdesc_tab[3];

	res =funionfs_realpath(path, &getpath, &setpath, &newpath, pdesc_tab);
	if (res < 0)
		return res;

	DEBUG_PRINTF("GETATTR(%s,%s,%s): %d \n", getpath, setpath, newpath,size);

	if (res & FILE_EXIST)
		res = readlink((const char *)getpath, buf, size - 1);
	else {
		res=-1;
		errno = ENOENT;
	}

// ATTENTION, il faut corriger le chemin afin de prendre en compte les chemins relalifs de faon correcte !
// MAJ: A voir, peut etre que la reecriture doit tre optionnelle

	free(getpath);
	free(setpath);
	free(newpath);

	if(res < 0)
		return -errno;

	buf[res] = '\0';
	return 0;
}


static int funionfs_mknod(const char *path, mode_t mode, dev_t dev)
{
	char *getpath, *setpath, *newpath;
	int res, status;
	int len;
	struct unionfs_desc *pdesc_tab[3];

	status =funionfs_realpath(path, &getpath, &setpath, &newpath, pdesc_tab);
	if (status < 0)
		return status;

	DEBUG_PRINTF("MKNOD(%s,%s,%s): %8.8x %8.8x\n", getpath, setpath, newpath, (unsigned int)mode, (unsigned int)dev);

	res = -1;

	do {
		if (status & FILE_EXIST) {
			errno = EEXIST;
			break;
		}
		if (! (status & FILE_CREATABLE)) {
			errno = EROFS;
			break;
		}
		if (status & FILE_DELETED) {
			len = strlen(newpath);
			strcat(newpath, funionfs.del_string);
			res=unlink(newpath);
			if (res == -1) {
				break;
			}
			newpath[len]='\0';
		} else {
			res=funionfs_pathtowrite(path);
			if (res < 0)
				break;
		}

		if (S_ISFIFO(mode))
			res = mkfifo(newpath, mode);
		else
			res = mknod(newpath, mode, dev);
		if((res == 0) && (geteuid() == 0 )) {
			res = chown(newpath,fuse_get_context()->uid, fuse_get_context()->gid);
		}

	} while(0);

	free(getpath);
	free(setpath);
	free(newpath);

	if(res == -1)
		return -errno;
	return 0;
}

static int funionfs_mkdir(const char *path, mode_t mode)
{
	char *getpath, *setpath, *newpath;
	int res, status;
	int len;
	struct unionfs_desc *pdesc_tab[3];

	status =funionfs_realpath(path, &getpath, &setpath, &newpath, pdesc_tab);
	if (status < 0)
		return status;

	DEBUG_PRINTF("MKDIR(%s,%s,%s): %8.8x \n", getpath, setpath, newpath, mode);

	res = -1;

	do {
		if (! (status & FILE_CREATABLE)) {
			errno = EROFS;
			break;
		}
		if (status & FILE_DELETED) {
			len = strlen(newpath);
			strcat(newpath, funionfs.del_string);
			res=unlink(newpath);
			if (res == -1) {
				break;
			}
			newpath[len]='\0';
		} else if (status & FILE_EXIST) {
			errno = EEXIST;
			break;
		} else {
			res=funionfs_pathtowrite(path);
			if (res < 0)
				break;
		}

		res = mkdir(newpath, mode);
		if((! res) && (!geteuid())) {
			res = chown(newpath,fuse_get_context()->uid, fuse_get_context()->gid);
		}

	} while(0);

	free(getpath);
	free(setpath);
	free(newpath);

	if(res < 0)
		return -errno;
	return 0;
}

static int funionfs_unlink(const char *path)
{
	char *getpath, *setpath, *newpath;
	int status, res;
	struct unionfs_desc *pdesc_tab[3];

	status =funionfs_realpath(path, &getpath, &setpath, &newpath, pdesc_tab);
	if (status < 0)
		return status;

	DEBUG_PRINTF("UNLINK(%s,%s,%s) \n", getpath, setpath, newpath);

	res=-1;
	do {

		if (! (status & FILE_CREATABLE)) {
			errno=EROFS;
			break;
		}
		if ((! (status & FILE_EXIST)) || (status & FILE_DELETED)) {
			errno=ENOENT;
			break;
		}
		if (status & FILE_WRITEABLE) {
			res = unlink(setpath);
			if (res < 0)
				break;
			if (status & FILE_TODELETE) {
				strcat(newpath, funionfs.del_string);
				res=funionfs_pathtowrite(newpath);
				if (res < 0)
					break;
				res = mknod(newpath, 0777, 0);
			}
		} else {
			strcat(newpath, funionfs.del_string);
			res=funionfs_pathtowrite(newpath);
			if (res < 0)
				break;
			res = mknod(newpath, 0777, 0);
		}
	} while(0);

	free(getpath);
	free(setpath);
	free(newpath);

	if(res < 0 )
		return -errno;
	return 0;

}

static int funionfs_rmdir(const char *path)
{
	char *getpath, *setpath, *newpath;
	int res, status;
	struct unionfs_desc *pdesc_tab[3];

	status =funionfs_realpath(path, &getpath, &setpath, &newpath, pdesc_tab);
	if (status < 0)
		return status;

	DEBUG_PRINTF("RMDIR(%s,%s,%s) \n", getpath, setpath, newpath);

	res=-1;
	do {
		if (! (status & FILE_CREATABLE)) {
			errno=EROFS;
			break;
		}
		if ((! (status & FILE_EXIST)) || (status & FILE_DELETED)) {
			errno=ENOENT;
			break;
		}
		if (status & FILE_WRITEABLE) {
			res = rmdir(setpath);
			if (res < 0)
				break;
			if (status & FILE_TODELETE) {
				strcat(newpath, funionfs.del_string);
				res = mknod(newpath, 0777, 0);
			}
		} else {
			strcat(newpath, funionfs.del_string);
			res = mknod(newpath, 0777, 0);
		}
	} while(0);

	free(getpath);
	free(setpath);
	free(newpath);

	if(res < 0)
		return -errno;
	return 0;
}

static int funionfs_symlink(const char *from, const char *to)
{
	char *getpath, *setpath, *newpath;
	int res, status;
	struct unionfs_desc *pdesc_tab[3];

	status =funionfs_realpath(to, &getpath, &setpath, &newpath, pdesc_tab);
	if (status < 0)
		return status;

	DEBUG_PRINTF("SYMLINK(%s,%s,%s): %s \n", getpath, setpath, newpath, from);

	res = -1;

	do {
		if (! (status & FILE_CREATABLE)) {
			errno = EROFS;
			break;
		}
		if (status & FILE_DELETED) {
			strcat(newpath, funionfs.del_string);
			res=unlink(newpath);
		} else if (status & FILE_EXIST) {
			errno = EEXIST;
			break;
		} else {
			res=funionfs_pathtowrite(to);
			if (res < 0)
				break;
		}

		res = symlink(from, newpath);
		if (res == -1) {
			break;
		}
		if((! res) && (!geteuid())) {
			res = lchown(newpath,fuse_get_context()->uid, fuse_get_context()->gid);
		}

	} while(0);

	free(getpath);
	free(setpath);
	free(newpath);

	if(res == -1)
		return -errno;
	return 0;

}

static int funionfs_rename(const char *from, const char *to)
{
	char *to_getpath, *to_setpath, *to_newpath;
	char *from_getpath, *from_setpath, *from_newpath;
	struct unionfs_desc *pdesc_tab_to[3];
	struct unionfs_desc *pdesc_tab_from[3];
	int res;
	int status1, status2;

	status1 =funionfs_realpath(to, &to_getpath, &to_setpath, &to_newpath, pdesc_tab_to);
	if (status1 < 0)
		return status1;

	status2 =funionfs_realpath(from, &from_getpath, &from_setpath, &from_newpath, pdesc_tab_from);
	if (status2 < 0) {
		free(to_getpath);
		free(to_setpath);
		free(to_newpath);
		return status2;
	}

	DEBUG_PRINTF("RENAME(%s,%s,%s): (%s,%s,%s) \n", to_getpath, to_setpath, to_newpath, from_getpath, from_setpath, from_newpath);
	do {
		/* test if destination directory exists and create if not */
		/* destination file is overwritten if exist */
		if (! (status1 & FILE_CREATABLE)) {
			errno = EROFS;
			break;
		}
		if (status1 & FILE_DELETED) {
			strcat(to_newpath, funionfs.del_string);
			res=unlink(to_newpath);
		} else {
			res=funionfs_pathtowrite(to);
			if (res < 0)
				break;
		}
		if (status2 & FILE_WRITEABLE) {
			res = rename(from_setpath, to_newpath);
			if (res < 0)
				break;
			if (status2 & FILE_TODELETE) {
				strcat(from_newpath, funionfs.del_string);
				res = mknod(from_newpath, 0777, 0);
			}
		} else {
			res = funionfs_copytowrite(from_getpath, from_newpath);
			if (res < 0)
				break;
			res = rename(from_newpath, to_newpath);
			if (res < 0)
				break;
			strcat(from_newpath, funionfs.del_string);
			res = mknod(from_newpath, 0777, 0);
		}

	} while(0);

	free(to_getpath);
	free(to_setpath);
	free(to_newpath);
	free(from_getpath);
	free(from_setpath);
	free(from_newpath);

	if(res == -1)
		return -errno;
	return 0;

}

static int funionfs_link(const char *from, const char *to)
{
	char *to_getpath, *to_setpath, *to_newpath;
	char *from_getpath, *from_setpath, *from_newpath;
	int res, status1,status2;
	struct unionfs_desc *pdesc_tab_to[3];
	struct unionfs_desc *pdesc_tab_from[3];

	status1 =funionfs_realpath(to, &to_getpath, &to_setpath, &to_newpath, pdesc_tab_to);
	if (status1 < 0)
		return status1;

	status2 =funionfs_realpath(from, &from_getpath, &from_setpath, &from_newpath, pdesc_tab_from);
	if (status2 < 0) {
		free(to_getpath);
		free(to_setpath);
		free(to_newpath);
		return status2;
	}

	DEBUG_PRINTF("LINK(%s,%s,%s): (%s,%s,%s) \n", to_getpath, to_setpath, to_newpath, from_getpath, from_setpath, from_newpath);

	res = -1;

	do {
		if (status1 & FILE_EXIST) {
			errno = EEXIST;
			break;
		}
		if (! (status1 & FILE_CREATABLE)) {
			errno = EROFS;
			break;
		}
		if (status1 & FILE_DELETED) {
			strcat(to_newpath, funionfs.del_string);
			res=unlink(to_newpath);
		} else {
			res=funionfs_pathtowrite(to);
			if (res < 0)
				break;
		}

		res = link(from_getpath, to_newpath);
		if (res == -1) {
			res = funionfs_copytowrite(from_getpath, from_newpath);
			if (res < 0)
				break;
			res = link(from_newpath, to_newpath);
			if (res == -1)
				break;
		}

	} while(0);

	free(to_getpath);
	free(to_setpath);
	free(to_newpath);
	free(from_getpath);
	free(from_setpath);
	free(from_newpath);

	if(res == -1)
		return -errno;
	return 0;
}


static int funionfs_chmod(const char *path, mode_t mode)
{
	int res;
	char *getpath, *setpath, *newpath;
	struct unionfs_desc *pdesc_tab[3];
	int status;

	status =funionfs_realpath(path, &getpath, &setpath, &newpath, pdesc_tab);
	if (status < 0)
		return status;

	DEBUG_PRINTF("CHMOD(%s,%s,%s): %x \n", getpath, setpath, newpath, mode);

	res = -1;

	do {
		if (*setpath == '\0') {
			res = funionfs_copytowrite(getpath, newpath);
			if (res < 0)
				break;
			res = chmod(newpath, mode);
		} else
			res = chmod(setpath, mode);

	} while(0);

	free(getpath);
	free(setpath);
	free(newpath);

	if(res == -1)
		return -errno;

	return 0;
}

static int funionfs_chown(const char *path, uid_t uid, gid_t gid)
{
	int res;
	char *getpath, *setpath, *newpath;
	struct unionfs_desc *pdesc_tab[3];
	int status;

	status =funionfs_realpath(path, &getpath, &setpath, &newpath, pdesc_tab);
	if (status < 0)
		return status;

	DEBUG_PRINTF("CHOWN(%s,%s,%s): %d %d \n", getpath, setpath, newpath, uid, gid);

	res = -1;

	do {
		if (*setpath == '\0') {
			res = funionfs_copytowrite(getpath, newpath);
			if (res < 0)
				break;
			res = lchown(newpath, uid, gid);
		} else
			res = lchown(setpath, uid, gid);

	} while(0);

	free(getpath);
	free(setpath);
	free(newpath);

	if(res == -1)
		return -errno;

	return 0;
}

static int funionfs_truncate(const char *path, off_t size)
{
	int res;
	char *getpath, *setpath, *newpath;
	struct unionfs_desc *pdesc_tab[3];
	int status;

	status =funionfs_realpath(path, &getpath, &setpath, &newpath, pdesc_tab);
	if (status < 0)
		return status;

	DEBUG_PRINTF("TRUNCATE(%s,%s,%s): %x \n", getpath, setpath, newpath, (unsigned int)size);

	res = -1;

	do {
		if (*setpath == '\0') {
			res = funionfs_copytowrite(getpath, newpath);
			if (res < 0)
				break;
			res = truncate(newpath, size);
		} else
			res = truncate(setpath, size);

	} while(0);

	free(getpath);
	free(setpath);
	free(newpath);

	if(res == -1)
		return -errno;

	return 0;
}

static int funionfs_utime(const char *path, struct utimbuf *buf)
{
	int res;
	char *getpath, *setpath, *newpath;
	struct unionfs_desc *pdesc_tab[3];
	int status;

	status =funionfs_realpath(path, &getpath, &setpath, &newpath, pdesc_tab);
	if (status < 0)
		return status;

	DEBUG_PRINTF("UTIME(%s,%s,%s) \n", getpath, setpath, newpath);

	res = -1;

	do {
		if (*setpath == '\0') {
			res = funionfs_copytowrite(getpath, newpath);
			if (res < 0)
				break;
			res = utime(newpath, buf);
		} else
			res = utime(setpath, buf);

	} while(0);

	free(getpath);
	free(setpath);
	free(newpath);

	if(res == -1)
		return -errno;

	return 0;
}

static int funionfs_open(const char *path, struct fuse_file_info *info)
{
	int res, status;

	char *getpath, *setpath, *newpath;
	struct unionfs_desc *pdesc_tab[3];

	status =funionfs_realpath(path, &getpath, &setpath, &newpath, pdesc_tab);
	if (status < 0)
		return status;

	DEBUG_PRINTF("OPEN(%s,%s,%s)\n", getpath, setpath, newpath);

	res = -1;

	do {
		if ((! (status & FILE_EXIST)) || (status & FILE_DELETED)) {
			errno=ENOENT;
			break;
		}
#if FUSE_USE_VERSION >= 25
		res = open(getpath, info->flags);
#else
		res = open(getpath, O_RDONLY);
#endif
		if (res < 0)
			break;
		close(res);

	} while(0);

	free(getpath);
	free(setpath);
	free(newpath);

	if (res < 0) 
		return -errno;

	return 0;

}

static int funionfs_read(const char *path, char *buf, size_t size, off_t offset, struct fuse_file_info *info)
{
	int fd;
	int res, status;

	char *getpath, *setpath, *newpath;
	struct unionfs_desc *pdesc_tab[3];

	(void)info;

	status =funionfs_realpath(path, &getpath, &setpath, &newpath, pdesc_tab);
	if (status < 0)
		return status;

	DEBUG_PRINTF("READ(%s,%s,%s): %x %x\n", getpath, setpath, newpath, (unsigned int)offset, (unsigned int)size);

	res = -1;

	do {
		if ((! (status & FILE_EXIST)) || (status & FILE_DELETED)) {
			errno=ENOENT;
			break;
		}
		fd = open(getpath, O_RDONLY);
		res = 0;
	} while(0);

	free(getpath);
	free(setpath);
	free(newpath);

	if(res < 0)
		return -errno;
	if(fd < 0)
		return -errno;

	res = lseek(fd, offset, SEEK_SET);
	if(res < 0)
		res = -errno;
	else {
		res = read(fd, buf, size);
		if(res < 0)
			res = -errno;
	}

	close(fd);

	if (res < 0) 
		return -errno;

	return res;
}

/* liste un repertoire
 * Cet algorithme est simpliste et mauvais. les performances sont dplorables
 * sur des rpertoires gigantesques.  revoir avec un algorithme de hashing
 * ou avec une recherche logarithmique.
 */

static int funionfs_readdir(const char *path, void *buf, fuse_fill_dir_t filler,
                         off_t offset, struct fuse_file_info *fi)
{
	struct unionfs_dname {
		char *name;
		struct unionfs_dname *pnext;
	};

	DIR *dir;
	struct dirent *curdir;
	char *getpath;
	char *curcpy;
	struct unionfs_walk *pwalk;
	int len;
	long type;

	struct unionfs_dname *pfirst, *pcur, *pprev;

	(void)offset;
	(void)fi;

	pwalk =funionfs_openpath(path);
	if (pwalk  == NULL)
		return -errno;

	DEBUG_PRINTF("READDIR(%s)\n", path);

	pfirst = NULL;

	while(1) {
		getpath = funionfs_readpath(pwalk, &type);
		if (getpath == NULL)
			break;

		dir = opendir(getpath);

		if (dir == NULL)
			return -errno;

		while(1) {
			curdir = readdir(dir);
			if (curdir == NULL)
				break;

			if (getrightstr(curdir->d_name, funionfs.del_string) != 0) {

				len = strlen(curdir->d_name) - funionfs.del_len;
				curcpy = (char *)malloc(len + 1);
				if (curcpy == NULL) {
					closedir(dir);
					funionfs_closepath(pwalk);
					/* free de la liste des reps */
					pcur = pfirst;
					while(pcur != NULL) {
						pprev = pcur;
						pcur = pcur->pnext;
						free(pprev->name);
						free(pprev);
					}
					return -errno;
				}
				strncpy(curcpy, curdir->d_name, len);
				curcpy[len]='\0';

				pcur = pfirst;
				pprev = NULL;
				/* chercher le repertoire et l'effacer de la liste si prsent */
				while(pcur != NULL) {
					if (strcmp(curcpy, pcur->name) == 0) {
						free(pcur->name);
						if (pprev == NULL)
							pfirst = pcur->pnext;
						else 
							pprev->pnext = pcur->pnext;
						free(pcur);
						break;
					}
					pprev = pcur;
					pcur = pcur->pnext;
				}
				free(curcpy);
			} else {

				pcur = pfirst;
				/* chercher le repertoire ne l'insrer que si absent */
				while(pcur != NULL) {
					if (strcmp(curdir->d_name, pcur->name) == 0) {
						break;
					}
					pcur = pcur->pnext;
				}
				if (pcur == NULL) {
					pcur = (struct unionfs_dname *) malloc(sizeof(struct unionfs_dname));
					if (pcur == NULL) {
						closedir(dir);
						funionfs_closepath(pwalk);
						/* free de la liste des reps */
						pcur = pfirst;
						while(pcur != NULL) {
							pprev = pcur;
							pcur = pcur->pnext;
							free(pprev->name);
							free(pprev);
						}
						return -errno;
					}
					pcur->name = strdup(curdir->d_name);
					if (pcur->name == NULL) {
						free(pcur);
						closedir(dir);
						funionfs_closepath(pwalk);
						/* free de la liste des reps */
						pcur = pfirst;
						while(pcur != NULL) {
							pprev = pcur;
							pcur = pcur->pnext;
							free(pprev->name);
							free(pprev);
						}
						return -errno;
					}
					pcur->pnext = pfirst;
					pfirst = pcur;
				}
			}
		}
		closedir(dir);
	}
	funionfs_closepath(pwalk);

	pcur = pfirst;
	while(pcur != NULL) {
		filler(buf, pcur->name, NULL, 0);
		pprev = pcur;
		pcur = pcur->pnext;
		free(pprev->name);
		free(pprev);
	}
	
	return 0;
}

static int funionfs_write(const char *path, const char *buf, size_t size, off_t offset, struct fuse_file_info *info)
{
	int res;
	char *getpath, *setpath, *newpath;
	struct unionfs_desc *pdesc_tab[3];
	int status;
	int fd;

	(void)info;

	status =funionfs_realpath(path, &getpath, &setpath, &newpath, pdesc_tab);
	if (status < 0)
		return status;

	DEBUG_PRINTF("WRITE(%s,%s,%s): %x %x\n", getpath, setpath, newpath, (unsigned int)offset, (unsigned int)size);

	res = -1;

	do {
		if ((! (status & FILE_EXIST)) || (status & FILE_DELETED)) {
			errno=ENOENT;
			break;
		}
		if (! (status & FILE_CREATABLE)) {
			errno=EROFS;
			break;
		}
		if (*setpath == '\0') {
			res = funionfs_copytowrite(getpath, newpath);
			if (res < 0)
				break;
			fd = open(newpath, O_WRONLY);
		} else {
			res=0;
			fd = open(setpath, O_WRONLY);
		}

	} while(0);

	free(getpath);
	free(setpath);
	free(newpath);

	if(res < 0)
		return -errno;
	if(fd < 0)
		return -errno;

	res = lseek(fd, offset, SEEK_SET);
	if(res < 0)
		res = -errno;
	else {
		res = write(fd, buf, size);
		if(res < 0)
			res = -errno;
	}

	close(fd);
	return res;
}

#if FUSE_USE_VERSION == 25
static int funionfs_statfs(const char *path, struct statvfs *fst)
{
	int rv;
	char *getpath, *setpath, *newpath;
	int status;
	struct unionfs_desc *pdesc_tab[3];

	status =funionfs_realpath(path, &getpath, &setpath, &newpath, pdesc_tab);
	if (status < 0)
		return status;

	DEBUG_PRINTF("STATFS(%s,%s,%s) \n", getpath, setpath, newpath);

	rv = statvfs(newpath,fst);

	return rv;
}

#else

static int funionfs_statfs(const char *path, struct statfs *fst)
{
	int rv;
	char *getpath, *setpath, *newpath;
	int status;
	struct unionfs_desc *pdesc_tab[3];

	status =funionfs_realpath(path, &getpath, &setpath, &newpath, pdesc_tab);
	if (status < 0)
		return status;

	DEBUG_PRINTF("STATFS(%s,%s,%s) \n", getpath, setpath, newpath);

	rv = statfs(newpath,fst);

	return rv;
}

#endif

#ifdef HAVE_SETXATTR

static int funionfs_setxattr(const char *path, const char *name, const char *value, size_t size, int flags)
{
	int res;
	char *getpath, *setpath, *newpath;
	struct unionfs_desc *pdesc_tab[3];
	int status;

	status =funionfs_realpath(path, &getpath, &setpath, &newpath, pdesc_tab);
	if (status < 0)
		return status;

	DEBUG_PRINTF("SETXATTR(%s,%s,%s): %s, %s, %d, %d\n", getpath, setpath, newpath, name, value, size, flags);

	res = -1;

	do {
		if (*setpath == '\0') {
			res = funionfs_copytowrite(getpath, newpath);
			if (res < 0)
				break;
			res = lsetxattr(newpath, name, value, size, flags);
		} else
			res = lsetxattr(setpath, name, value, size, flags);

	} while(0);

	free(getpath);
	free(setpath);
	free(newpath);

	if(res == -1)
		return -errno;

	return 0;
}

static int funionfs_getxattr(const char *path, const char *name, char *value, size_t size)
{
	int res = 0;
	struct unionfs_desc *pdesc_tab[3];

	char *getpath, *setpath, *newpath;

	res =funionfs_realpath(path, &getpath, &setpath, &newpath, pdesc_tab);
	if (res < 0)
		return res;

	DEBUG_PRINTF("GETXATTR(%s,%s,%s): %s, %d\n", getpath, setpath, newpath, name, size);

	if (res & FILE_EXIST) {
		res = lgetxattr(getpath, name, value, size);
	} else {
		res = -1;
		errno=ENOENT;
	}

	free(getpath);
	free(setpath);
	free(newpath);

	if (res < 0) 
		return -errno;

	return res;
}

static int funionfs_listxattr(const char *path, char *list, size_t size)
{
	int res = 0;
	struct unionfs_desc *pdesc_tab[3];

	char *getpath, *setpath, *newpath;

	res =funionfs_realpath(path, &getpath, &setpath, &newpath, pdesc_tab);
	if (res < 0)
		return res;

	DEBUG_PRINTF("LISTXATTR(%s,%s,%s): %d\n", getpath, setpath, newpath, size);

	if (res & FILE_EXIST) {
		res = llistxattr(getpath, list, size);
	} else {
		res = -1;
		errno=ENOENT;
	}

	free(getpath);
	free(setpath);
	free(newpath);

	if (res < 0) 
		return -errno;

	return res;
}

static int funionfs_removexattr(const char *path, const char *name)
{
	int res;
	char *getpath, *setpath, *newpath;
	struct unionfs_desc *pdesc_tab[3];
	int status;

	status =funionfs_realpath(path, &getpath, &setpath, &newpath, pdesc_tab);
	if (status < 0)
		return status;

	DEBUG_PRINTF("SETXATTR(%s,%s,%s): %s\n", getpath, setpath, newpath, name);

	res = -1;

	do {
		if (*setpath == '\0') {
			res = funionfs_copytowrite(getpath, newpath);
			if (res < 0)
				break;
			res = lremovexattr(newpath, name);
		} else
			res = lremovexattr(setpath, name);

	} while(0);

	free(getpath);
	free(setpath);
	free(newpath);

	if(res == -1)
		return -errno;

	return 0;

}
#endif /* HAVE_SETXATTR */

#if FUSE_USE_VERSION >= 25

static int funionfs_create(const char *path, mode_t mode, struct fuse_file_info *fi)
{
	int res, status, len;
	char *getpath, *setpath, *newpath;
	struct unionfs_desc *pdesc_tab[3];

	status =funionfs_realpath(path, &getpath, &setpath, &newpath, pdesc_tab);
	if (status < 0)
		return status;

	DEBUG_PRINTF("CREATE(%s,%s,%s)\n", getpath, setpath, newpath);

	res = -1;

	do {
		if (status & FILE_EXIST) {
			errno = EEXIST;
			break;
		}
		if (! (status & FILE_CREATABLE)) {
			errno = EROFS;
			break;
		}
		if (status & FILE_DELETED) {
			len = strlen(newpath);
			strcat(newpath, funionfs.del_string);
			res=unlink(newpath);
			if (res == -1) {
				break;
			}
			newpath[len]='\0';
		} else {
			res=funionfs_pathtowrite(path);
			if (res < 0)
				break;
		}

		res = open(newpath, fi->flags, mode);
		if((res != -1) && (geteuid() == 0 )) {
			close(res);
			res = chown(newpath,fuse_get_context()->uid, fuse_get_context()->gid);
		}

	} while(0);

	free(getpath);
	free(setpath);
	free(newpath);

	if (res < 0) 
		return -errno;

	return 0;
}

static int funionfs_access(const char *path, int mask)
{
	int res, status;
	char *getpath, *setpath, *newpath;
	struct unionfs_desc *pdesc_tab[3];

	status =funionfs_realpath(path, &getpath, &setpath, &newpath, pdesc_tab);
	if (status < 0)
		return status;

	DEBUG_PRINTF("ACCESS(%s,%s,%s)\n", getpath, setpath, newpath);

	res = -1;

	do {
		if ((! (status & FILE_EXIST)) || (status & FILE_DELETED)) {
			errno=ENOENT;
			break;
		}
		res = access(getpath, mask);
		if (res < 0)
			break;
	} while(0);

	free(getpath);
	free(setpath);
	free(newpath);

	if (res < 0) 
		return -errno;

	return 0;
}

#endif

void funionfs_destroy(void *nouse)
{
	struct unionfs_desc *pdesc, *pnext;

	(void)nouse;

	for (pdesc = unionfs_list; pdesc;)
	{
		free(pdesc->path);
		pnext = pdesc->pnext;
		free(pdesc);
		pdesc = pnext;
	}
	unionfs_list = NULL;
}

static struct fuse_operations funionfs_oper = {
	.getattr	= funionfs_getattr,
	.readlink	= funionfs_readlink,
	.mknod		= funionfs_mknod,
	.mkdir		= funionfs_mkdir,
	.unlink		= funionfs_unlink,
	.rmdir		= funionfs_rmdir,
	.symlink	= funionfs_symlink,
	.rename		= funionfs_rename,
	.link		= funionfs_link,
	.chmod		= funionfs_chmod,
	.chown		= funionfs_chown,
	.truncate	= funionfs_truncate,
	.utime		= funionfs_utime,
	.open		= funionfs_open,
	.read		= funionfs_read,
	.readdir	= funionfs_readdir,
	.write		= funionfs_write,
	.statfs		= funionfs_statfs,
#ifdef HAVE_SETXATTR
	.setxattr	= funionfs_setxattr,
	.getxattr	= funionfs_getxattr,
	.listxattr	= funionfs_listxattr,
	.removexattr	= funionfs_removexattr,
#endif
#if FUSE_USE_VERSION >= 25
	.create		= funionfs_create,
	.access		= funionfs_access,
#endif
// 	.init		= funionfs_init,
	.destroy	= funionfs_destroy,
};

#define FUNIONFS_OPT(t, p, v) { t, offsetof(struct funionfs, p), v }

enum {
	KEY_VERSION,
	KEY_HELP
};

static struct fuse_opt funionfs_opts[] = {
    FUNIONFS_OPT("dirs=%s",		dirlist, 0),
    FUNIONFS_OPT("copyupuid=%s",	uid, 0),
    FUNIONFS_OPT("copyupgid=%s",	gid, 0),
    FUNIONFS_OPT("copyupmode=%s",	mode, 0),
    FUNIONFS_OPT("del_string=%s",	del_string, 0),
    FUNIONFS_OPT("delete=all",		delete, 0),
    FUNIONFS_OPT("delete=whiteout",	delete, 1),
    FUNIONFS_OPT("copyup=preserve",	copyup, 0),
    FUNIONFS_OPT("copyup=currentuser",	copyup, 1),
    FUNIONFS_OPT("copyup=mounter",	copyup, 2),

    FUSE_OPT_KEY("-V",             KEY_VERSION),
    FUSE_OPT_KEY("--version",      KEY_VERSION),
    FUSE_OPT_KEY("-h",             KEY_HELP),
    FUSE_OPT_KEY("--help",         KEY_HELP),
    FUSE_OPT_END
};

static void usage(const char *progname)
{
    fprintf(stderr,
"usage: %s [ /upper_rwdir ] mountpoint [options]\n"
"\n"
"general options:\n"
"    -o opt,[opt...]        mount options\n"
"    -h   --help            print help\n"
"    -V   --version         print version\n"
"\n"
"FUNIONFS options:\n"
"    -o funionfs_debug      print some debugging information\n"
"    -o dirs=LIST           A colon separated list of directories\n"
"             /branch1=rw      a directory in a read-write mode\n"
"             /branch1         if no permission is specified, the branch\n"
"                              is in a read-write mode\n"
"             /branch2=ro      a directory in a read-only mode\n"
"    -o delete=TYPE         specifies how funionfs deletes and renames\n"
"             all              delete all objects in read-write branch (default)\n"
"                              and add a whiteout file if objects exists in\n"
"                              read-only branches.\n"
"             whiteout         only add a whiteout file in upper (rw) branch\n"
"                              or delete file in upper branch.\n"
"    -o copyup=TYPE         specifies how to manage attributes\n"
"             preserve         preserve uid, gid and mode of object (default)\n"
"             currentuser      the object owns to the user doing the operation\n"
"             mounter          the object owns to the mounter\n"
"    -o copyupuid=N         uid of the copied-up files (no default)\n"
"    -o copyupgid=N         gid of the copied-up files (no default)\n"
"    -o copyupmode=N        mode of the copied-up files (no default)\n"
"    -o del_string=N        string to mark a deleted file(default: %s )\n"
"\n"
"You could use the mount command instead :\n"
"   mount -t fuse funionfs#/upper_rwdir mountpoint [ options ]\n"
"\n"
"There is no implementation of -o delete and -o copyup for now.\n"
"\n", progname, DELETED_STRING);
}

static int funionfs_opt_proc(void *data, const char *arg, int key,
                          struct fuse_args *outargs)
{
	(void) data;

	switch (key) {
	case FUSE_OPT_KEY_NONOPT:
		if (*funionfs.firstdir == '\0') {
			funionfs.firstdir = strdup(arg);
			return 0;
		} else if (*funionfs.mountpoint == '\0')
			funionfs.mountpoint = strdup(arg);
		return 1;
	case FUSE_OPT_KEY_OPT:
		if (!strcmp(arg, "-d")) {
			funionfs.debug = 1;
		}
		return 1;
	case KEY_HELP:
		usage(outargs->argv[0]);
		fuse_opt_add_arg(outargs, "-ho");
		fuse_main(outargs->argc, outargs->argv, &funionfs_oper);
		exit(1);

	case KEY_VERSION:

		fprintf(stderr, "FUNIONFS version %s\n", VERSION);
#if FUSE_VERSION >= 25
		fuse_opt_add_arg(outargs, "--version");
		fuse_main(outargs->argc, outargs->argv, &funionfs_oper);
#else
		printf("FUSE VERSION: %d.%d\n", FUSE_MAJOR_VERSION, FUSE_MINOR_VERSION);
#endif
		exit(0);

	default:
		fprintf(stderr, "internal error: %d\n", key);
		abort();
	}
}

	int
parse_and_add_dirs(void)
{
	char *ptr, *end;
	char *path;
	char basepath[4096];
	int type, ret;

	ptr = getcwd(basepath, 4096);
	if (ptr == NULL)
		return -errno;

	// first add dirlist to the hierarchy

	ptr = funionfs.dirlist;
	while (*ptr)
	{
		end = ptr;
		while (*end && (*end != ':') && (*end != '=')) {
			end++;
		}

		type = 0; // default type is Read/Write

		if (*end == '=') {
			*end++='\0';
			if ((*end != 'r') && (*end != 'R'))
				return -1;
			end++;
			switch(*end++) {
				case 'o': case 'O':
					type = 1;
					break;
				case 'w': case 'W':
					break;
				default:
					return -1;
			}
			if (*end != ':' && *end)
				return -1;
		}
		if (*end == ':')
			*end++='\0';

		path = rel2abspath(ptr, basepath);
		ret = funionfs_addpath (path, type);
		free(path);
		if (ret < 0) {
			return -ret;
		}
		ptr = end;
	}

	// then add the main path to the list

	if (*funionfs.firstdir != '\0') {
		if (strcmp(funionfs.firstdir,"NONE") && strcmp(funionfs.firstdir, "none") && strcmp(funionfs.firstdir,"None")) {
			path = rel2abspath(funionfs.firstdir, basepath);
			ret = funionfs_addpath (path, type);
			free(path);
			if (ret < 0) {
				return -ret;
			}
		}
	}

	return 0;
}

int treat_options()
{
	if (!funionfs.del_string)
		return -1;
	funionfs.del_len = strlen(funionfs.del_string);

	return 0;
}


int main(int argc, char *argv[])
{
	int res;
	struct fuse_args args = FUSE_ARGS_INIT(argc, argv);

	if (fuse_opt_parse(&args, &funionfs, funionfs_opts, funionfs_opt_proc) == -1)
		exit(1);
	if (parse_and_add_dirs() < 0 )
		exit(1);
	if (treat_options() < 0 )
		exit(1);
	res = fuse_main(args.argc, args.argv, &funionfs_oper);
	fuse_opt_free_args(&args);

	return res;
}
